package ru.dbotthepony.kstarbound.world

import com.google.common.collect.ImmutableList
import it.unimi.dsi.fastutil.ints.IntList
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ReferenceLinkedOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.util.MailboxExecutorService
import ru.dbotthepony.kstarbound.util.ParallelPerform
import ru.dbotthepony.kstarbound.world.api.ICellAccess
import ru.dbotthepony.kstarbound.world.api.AbstractCell
import ru.dbotthepony.kstarbound.world.api.TileView
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kstarbound.world.entities.WorldObject
import ru.dbotthepony.kstarbound.world.physics.CollisionPoly
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import ru.dbotthepony.kstarbound.world.physics.Poly
import ru.dbotthepony.kstarbound.world.physics.getBlockPlatforms
import ru.dbotthepony.kstarbound.world.physics.getBlocksMarchingSquares
import ru.dbotthepony.kvector.api.IStruct2d
import ru.dbotthepony.kvector.api.IStruct2i
import ru.dbotthepony.kvector.arrays.Object2DArray
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.util2d.AABBi
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.Vector2i
import java.util.concurrent.ForkJoinPool
import java.util.concurrent.locks.ReentrantLock
import java.util.random.RandomGenerator

abstract class World<This : World<This, ChunkType>, ChunkType : Chunk<This, ChunkType>>(
	val seed: Long,
	val size: Vector2i,
	val loopX: Boolean,
	val loopY: Boolean
) : ICellAccess {
	val x: CoordinateMapper = if (loopX) CoordinateMapper.Wrapper(size.x) else CoordinateMapper.Clamper(size.x)
	val y: CoordinateMapper = if (loopY) CoordinateMapper.Wrapper(size.y) else CoordinateMapper.Clamper(size.y)

	// whenever provided cell position is within actual world borders, ignoring wrapping
	fun inBounds(x: Int, y: Int) = this.x.inBoundsCell(x) && this.y.inBoundsCell(y)
	fun inBounds(value: IStruct2i) = this.x.inBoundsCell(value.component1()) && this.y.inBoundsCell(value.component2())

	fun chunkFromCell(x: Int, y: Int) = ChunkPos(this.x.chunkFromCell(x), this.y.chunkFromCell(y))
	fun chunkFromCell(x: Double, y: Double) = ChunkPos(this.x.chunkFromCell(x.toInt()), this.y.chunkFromCell(y.toInt()))
	fun chunkFromCell(value: IStruct2i) = chunkFromCell(value.component1(), value.component2())
	fun chunkFromCell(value: IStruct2d) = chunkFromCell(value.component1(), value.component2())

	val background = TileView.Background(this)
	val foreground = TileView.Foreground(this)
	val mailbox = MailboxExecutorService()

	final override fun randomLongFor(x: Int, y: Int) = super.randomLongFor(x, y) xor seed

	override fun getCellDirect(x: Int, y: Int): AbstractCell {
		if (!this.x.inBoundsCell(x) || !this.y.inBoundsCell(y)) return AbstractCell.NULL
		return getCell(x, y)
	}

	override fun getCell(x: Int, y: Int): AbstractCell {
		return chunkMap.getCell(x, y)
	}

	override fun setCell(x: Int, y: Int, cell: AbstractCell): Boolean {
		return chunkMap.setCell(x, y, cell)
	}

	abstract inner class ChunkMap {
		abstract operator fun get(x: Int, y: Int): ChunkType?
		abstract fun compute(x: Int, y: Int): ChunkType?

		abstract fun remove(x: Int, y: Int)

		abstract fun getCell(x: Int, y: Int): AbstractCell
		abstract fun setCell(x: Int, y: Int, cell: AbstractCell): Boolean

		operator fun get(pos: ChunkPos) = get(pos.x, pos.y)

		protected fun create(x: Int, y: Int): ChunkType {
			val pos = ChunkPos(x, y)
			val chunk = chunkFactory(pos)
			val orphanedInThisChunk = ArrayList<Entity>()

			for (ent in orphanedEntities) {
				val (ex, ey) = ent.position

				if (this@World.x.chunkFromCell(ex) == x && this@World.y.chunkFromCell(ey) == y) {
					orphanedInThisChunk.add(ent)
				}
			}

			for (ent in orphanedInThisChunk) {
				ent.chunk = chunk
			}

			return chunk
		}
	}

	// around 30% slower than ArrayChunkMap, but can support insanely large worlds
	inner class SparseChunkMap : ChunkMap() {
		private val map = Long2ObjectOpenHashMap<ChunkType>()

		override fun getCell(x: Int, y: Int): AbstractCell {
			if (!this@World.x.isValidCellIndex(x) || !this@World.y.isValidCellIndex(y)) return AbstractCell.NULL
			val ix = this@World.x.cell(x)
			val iy = this@World.y.cell(y)
			return this[this@World.x.chunkFromCell(ix), this@World.y.chunkFromCell(iy)]?.getCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK) ?: AbstractCell.NULL
		}

		override fun get(x: Int, y: Int): ChunkType? {
			if (!this@World.x.inBoundsChunk(x) || !this@World.y.inBoundsChunk(y)) return null
			return map[ChunkPos.toLong(x, y)]
		}

		override fun compute(x: Int, y: Int): ChunkType? {
			if (!this@World.x.inBoundsChunk(x) || !this@World.y.inBoundsChunk(y)) return null
			return map[ChunkPos.toLong(x, y)] ?: create(x, y).also { map[ChunkPos.toLong(x, y)] = it }
		}

		override fun setCell(x: Int, y: Int, cell: AbstractCell): Boolean {
			if (!this@World.x.isValidCellIndex(x) || !this@World.y.isValidCellIndex(y)) return false
			val ix = this@World.x.cell(x)
			val iy = this@World.y.cell(y)
			val cx = this@World.x.chunkFromCell(ix)
			val cy = this@World.y.chunkFromCell(iy)
			return (map[ChunkPos.toLong(cx, cy)] ?: create(cx, cy).also { map[ChunkPos.toLong(cx, cy)] = it }).setCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK, cell)
		}

		override fun remove(x: Int, y: Int) {
			map.remove(ChunkPos.toLong(this@World.x.chunk(x), this@World.y.chunk(y)))
		}
	}

	inner class ArrayChunkMap : ChunkMap() {
		private val map = Object2DArray.nulls<ChunkType>(divideUp(size.x, CHUNK_SIZE), divideUp(size.y, CHUNK_SIZE))

		private fun getRaw(x: Int, y: Int): ChunkType {
			return map[x, y] ?: create(x, y).also { map[x, y] = it }
		}

		override fun compute(x: Int, y: Int): ChunkType? {
			if (!this@World.x.inBoundsChunk(x) || !this@World.y.inBoundsChunk(y)) return null
			return getRaw(x, y)
		}

		override fun getCell(x: Int, y: Int): AbstractCell {
			if (!this@World.x.isValidCellIndex(x) || !this@World.y.isValidCellIndex(y)) return AbstractCell.NULL
			val ix = this@World.x.cell(x)
			val iy = this@World.y.cell(y)
			return getRaw(ix ushr CHUNK_SIZE_BITS, iy ushr CHUNK_SIZE_BITS).getCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK)
		}

		override fun setCell(x: Int, y: Int, cell: AbstractCell): Boolean {
			if (!this@World.x.isValidCellIndex(x) || !this@World.y.isValidCellIndex(y)) return false
			val ix = this@World.x.cell(x)
			val iy = this@World.y.cell(y)
			return getRaw(ix ushr CHUNK_SIZE_BITS, iy ushr CHUNK_SIZE_BITS).setCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK, cell)
		}

		override fun get(x: Int, y: Int): ChunkType? {
			if (!this@World.x.inBoundsChunk(x) || !this@World.y.inBoundsChunk(y)) return null
			return getRaw(x, y)
		}

		override fun remove(x: Int, y: Int) {
			map[this@World.x.chunk(x), this@World.y.chunk(y)] = null
		}
	}

	val chunkMap: ChunkMap = if (size.x <= 32000 && size.y <= 32000) ArrayChunkMap() else SparseChunkMap()

	val random: RandomGenerator = RandomGenerator.of("Xoroshiro128PlusPlus")
	var gravity = Vector2d(0.0, -EARTH_FREEFALL_ACCELERATION)
	abstract val isClient: Boolean

	// used by world chunks to synchronize entity additions/transfer
	val entityListsLock = ReentrantLock()

	fun think() {
		try {
			mailbox.executeQueuedTasks()
			val entities = ObjectArrayList(entities)
			ForkJoinPool.commonPool().submit(ParallelPerform(entities.spliterator(), Entity::move)).join()
			mailbox.executeQueuedTasks()

			for (ent in entities) {
				ent.thinkShared()

				if (isClient)
					ent.thinkClient()
				else
					ent.thinkServer()
			}

			val objects = ObjectArrayList(objects)

			for (ent in objects) {
				ent.thinkShared()

				if (isClient)
					ent.thinkClient()
				else
					ent.thinkServer()
			}

			mailbox.executeQueuedTasks()
			thinkInner()
		} catch(err: Throwable) {
			throw RuntimeException("Ticking world $this", err)
		}
	}

	protected abstract fun thinkInner()

	val orphanedEntities = ReferenceOpenHashSet<Entity>()
	val entities = ReferenceLinkedOpenHashSet<Entity>()
	val objects = ReferenceLinkedOpenHashSet<WorldObject>()

	protected abstract fun chunkFactory(pos: ChunkPos): ChunkType

	fun queryCollisions(aabb: AABB): MutableList<CollisionPoly> {
		val result = ArrayList<CollisionPoly>()
		val tiles = aabb.encasingIntAABB()

		for (x in tiles.mins.x .. tiles.maxs.x) {
			for (y in tiles.mins.y .. tiles.maxs.y) {
				getBlocksMarchingSquares(x, y, foreground, CollisionType.DYNAMIC, result)
				getBlockPlatforms(x, y, foreground, CollisionType.PLATFORM, result)
			}
		}

		return result
	}
}
